@@ -26,6 +26,13 @@
 ##
 #############################################################################
 
+import os
+import sys
+
+sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
+from init_paths import init_test_paths
+init_test_paths(False)
+
 from PySide2 import QtScript
 
 
